/*
Arduino Mega Pro Mini 5v

To each 7 segment digit

 2 150 ohm - Segment A
 3 150 ohm - Segment B 
 4 150 ohm - Segment C
 5 150 ohm - Segment D
 6 150 ohm - Segment E
 7 150 ohm - Segment F
 8 150 ohm - Segment G
 9 150 ohm - Decimal point

Backplanes/digit grounds

10 Backplane digit 1 (Leftmost)
11 Backplane digit 2
12 Backplane digit 3
13 Backplane digit 4

Buttons:

14/A0 Menu button - GND
15/A1 Value button - GND

DS3231:

18/A4 SDA
19/A5 SCL

*/

#include "SevSeg.h"		// https://github.com/DeanIsMe/SevSeg
#include <Wire.h>
#include <DS3232RTC.h>  // http://github.com/JChristensen/DS3232RTC
#include <Time.h>    	// http://www.arduino.cc/playground/Code/Time
#include <Button.h>		// https://playground.arduino.cc/Code/Button
#include <EEPROM.h>

#define DEBUG

#ifdef DEBUG
	#define PRINTBEGIN() Serial.begin(9600); Serial.println(F("Serial initialised OK"))
	#define PRINTERROR(x) Serial.print(F("ERROR: ")); Serial.println(F(x))
	#define PRINTWARNING(x) Serial.print(F("WARNING: ")); Serial.println(F(x))
	#define PRINTVAR(x, y) Serial.print(F(x)); Serial.println(y)
	#define PRINT(x) Serial.print(x)
	#define PRINTF(x) Serial.print(F(x))
	#define PRINTLN(x) Serial.println(x)
	#define PRINTLNF(x) Serial.println(F(x))
#else
	#define PRINTBEGIN()
	#define PRINTERROR(x)
	#define PRINTWARNING(x)
	#define PRINTVAR(x, y)
	#define PRINT(x)
	#define PRINTF(x)
	#define PRINTLN(x)
	#define PRINTLNF(x)
#endif

// How long the temperature will be displayed
#define TEMP_TIMEOUT_MS 5000

// How long until the menu times out and goes back to clock
#define MENU_TIMEOUT_MS 5000

// How long to wait for a character to be scrolled
#define SCROLL_DELAY_MS 150

// How long to wait for a character to be scrolled 
#define SHORT_SCROLL_DELAY_MS 400

// How long to wait after value button has been pressed
#define VALUE_BUTTON_DELAY_MS 300

// How long to wait after menu button has been pressed
#define MENU_BUTTON_DELAY_MS 250

// How long to wait when screen is cleered
#define CLEAR_DELAY_MS		500;

// How long to wait to enter test mode
#define TEST_DISPLAY_DELAY_MS 1000

// How long to wait while testing the display
#define TEST_DISPLAY_DIGIT_DELAY 1000

// Address of initialsed flag in EEPROM
#define ADDR_FLAG		0

// Address of DST in EEPROM
#define ADDR_DST		1

// Address of seconds indicator in EEPROM
#define ADDR_SECONDS	2

// Address of 24 hour clock or not  in EEPROM
#define ADDR_FORMAT		3

// Value that means the EEPROM has been initialised
#define INITIALISED		0x99

// Number of digits in display
#define NUMBER_OF_DIGITS 4

// Macro to retrieve array length
#define ARRAY_LENGTH(x) (sizeof (x) / sizeof (x[0]))

// Enum to store page to display
enum page_t  {
	CLOCK,
	TEMPERATURE,
	HOUR,
	MINUTE,
	SECONDS,
	YEAR,
	MONTH,
	DAY,
	FORMAT
} page = CLOCK;

// Add an operator to allow the page enum to be incremented
inline page_t& operator++(page_t& p, int){
	const int i = static_cast<int>(p);
	p = static_cast<page_t>(i + 1);
	return p;
}

// Menu button
Button buttonMenu(14, PULLUP);

// Value button
Button buttonValue(15, PULLUP);

// Seven segment display
SevSeg display;

// Menu items
const char* prompt[] = {"HOUR", "MINUTE", "SECONDS", "YEAR", "MONTH", "DAY", "24" };

// Days in a month
uint8_t daysInMonth[] = {31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

// Display buffer
char buffer[NUMBER_OF_DIGITS + 2];

// Do we want the seconds to flash
bool showSeconds = true;

// Disable automatic DST adjustment
bool dstAdjusted = true;

// 24 hour by default
bool twentyFourHour = true;

// Setup!
void setup() {
	PRINTBEGIN();
	initEEPROM();
	initDisplay();
	initRTC();
	initTestDisplay();
}

// Secret demo mode for testing the screen (hold down any button during startup)
void initTestDisplay(){
	PRINTLNF("Waiting to enter display test..");
	unsigned long finish = millis() + TEST_DISPLAY_DELAY_MS;
	uint8_t i = 0;
	display.setChars("Test");
	while (millis() < finish){
		if (buttonMenu.isPressed() || buttonValue.isPressed()){
			PRINTLNF("Test entered");
			displayAndWait(TEST_DISPLAY_DELAY_MS);
			while (true){
				sprintf(buffer, "%d%d%d%d\0", i, i, i, i);
				display.setChars(buffer);
				i++;
				if (i == 10) i = 0;
				finish = millis() + TEST_DISPLAY_DIGIT_DELAY;
				while (millis() < finish){
					display.refreshDisplay();
					if (buttonMenu.uniquePress() || buttonValue.uniquePress()){
						PRINTLNF("Display test exited");
						clearAndWait();
						return;
					}
				} 
			}
		}
		display.refreshDisplay(); 
	}
	PRINTLNF("Display test skipped!");
	clearAndWait();
}

// Initialise the RTC
void initRTC(){
	PRINTLNF("Initialsing RTC...");
	setSyncProvider(RTC.get);
	if(timeStatus() != timeSet){
		PRINTWARNING("RTC fail!");
		scrollMessage("Problem with clock");
	}
	else{
		PRINTLNF("RTC OK!");
		PRINTVAR("DST? ", isDST());
		scrollMessage("Mase CLOCK");
	}
	displayAndWait(1000);
}

// Initialise the display
void initDisplay(){

	// https://github.com/DeanIsMe/SevSeg
	byte digitPins[] = {10, 11, 12, 13};
	byte segmentPins[] = {2, 3, 4, 5, 6, 7, 8, 9};
	bool resistorsOnSegments = true; // 'false' means resistors are on digit pins
	byte hardwareConfig = COMMON_ANODE; // Se README.md for options
	bool updateWithDelays = false; // Default. Recommended
	bool leadingZeros = false; // Use 'true' if you'd like to keep the leading zeros

	display.begin(hardwareConfig, NUMBER_OF_DIGITS, digitPins, segmentPins, resistorsOnSegments, updateWithDelays, leadingZeros);
	display.setBrightness(100);
}

// Main loop
void loop() {
	
	if ((page == CLOCK) || (page == TEMPERATURE)){

		if (page == CLOCK)
			displayClock();
		else if (page == TEMPERATURE)
			displayTemperature(0);

		// Go into setup pages
		if (buttonMenu.uniquePress()){
			displayAndWait(MENU_BUTTON_DELAY_MS);
			page = HOUR;
		}

		// Go into temperature mode
		if (buttonValue.uniquePress()){
			displayTemperature(1);
		}
	}
	else {

		while (page <= ARRAY_LENGTH(prompt) + 1){

			// Get initial value
			uint8_t value;

			switch (page){
				case HOUR:
					value = hour();
					break;
				case MINUTE:
					value = minute();
					break;
				case YEAR:
					value = (uint8_t) (year() - 2000);
					break;
				case MONTH:
					value = month();
					break;
				case DAY:
					value = day();
					break;
				case SECONDS:
					value = showSeconds;
					break;
				case FORMAT:
					value = twentyFourHour;
					break;
			}

			// Make prompt
			char p[3];
			strncpy(p, prompt[page - 2], 2);
			p[2] = 0;

			PRINTVAR("Menu page: ", prompt[page - 2]);

			unsigned long finish = millis() + MENU_TIMEOUT_MS;
			bool valueChanged = false;

			while (millis() < finish){
				sprintf(buffer, "%s%02d\0", p, value);
				display.setChars(buffer);
				display.refreshDisplay();

				if (buttonValue.isPressed()){
					value++;
					valueChanged = true;

					if ((page == HOUR) && (value > 23))
						value = 0;
					else if ((page == MINUTE) && (value > 59))
						value = 0;
					else if ((page == FORMAT) && (value > 1))
						value = 0;
					else if ((page == SECONDS) && (value > 1))
						value = 0;
					else if ((page == YEAR) && (value > 50))
						value = 18;
					else if ((page == MONTH) && (value > 12))
						value = 1;
					else if ((page == DAY) && (value > getDaysInMonth(month())))
						value = 1;

					sprintf(buffer, "%s%02d\0", p, value);
					display.setChars(buffer);
					display.refreshDisplay();

					displayAndWaitForButton(VALUE_BUTTON_DELAY_MS, buttonValue);

					finish = millis() + MENU_TIMEOUT_MS;
				}

				if (buttonMenu.uniquePress()){
					PRINTLNF("Next item!");
					displayAndWait(MENU_BUTTON_DELAY_MS);

					// Store value if it was changed
					if (valueChanged){
						if (page == HOUR){
							storeTime(value, minute(), second(), day(), month(), year());
							timeConfirmed();
						}
						else if (page == MINUTE)
							storeTime(hour(), value, 0, day(), month(), year());
						else if (page == SECONDS)
							saveShowSeconds(value);
						else if (page == FORMAT)
							saveFormat(value);	
						else if (page == YEAR)
							storeTime(hour(), minute(), second(), day(), month(), value);
						else if (page == MONTH) 
							storeTime(hour(), minute(), second(), day(), value, year());
						else if (page == DAY)
							storeTime(hour(), minute(), second(), value, month(), year());				
					}

					// Go to next page
					page++;
					break;
				}
			}

			// Menu item has timed out
			if (millis() >= finish){
				PRINTLNF("Menu timed out");
				break;
			}
			// If a value is changed then inform user its saved
			if (valueChanged){
				scrollMessage("Stored");
			}

		}
		
		// Exited so back to clock mode
		page = CLOCK;
	}
}

// Store any time settings
bool storeTime(uint8_t hour, uint8_t minute, uint8_t second, uint8_t day, uint8_t month, uint8_t year){
	PRINTLNF("RTC updated");
	setTime(hour, minute, second, day, month, year);
	return RTC.set(now());
}

// Display temperature
void displayTemperature(bool buttonPressed){
	static unsigned long finish;

	// Do we need to reset our timeout
	if (buttonPressed){
		finish = millis() + TEMP_TIMEOUT_MS;
		// Clear screen for a bit if we weren't on temperature page
		if (page != TEMPERATURE) clearAndWait();
	}

	// Display temperature for a certain time
	if (millis() < finish){
		// Convert float to char with 1 dp
		float temp = RTC.temperature() / 4.0f;
		dtostrf(temp, 3, 1, buffer);
		sprintf(buffer, "%sc\0", buffer);
		display.setChars(buffer);
		display.refreshDisplay();
		page = TEMPERATURE;
	}
	else {
		// Timeout so reset to clock mode
		page = CLOCK;
		clearAndWait();
	}
}

// Display clock
void displayClock(){
	static time_t oldTime;
	
	// Check we need to automatically adjust for DST
	if (checkDST())
		oldTime = 0; // Reset old time so new time is display straight away

	// If time has changed update display
	time_t currentTime = now();
	if (currentTime != oldTime){
		sprintf(buffer, "%02d%02d%s\0", hour() - (!twentyFourHour ? 12 : 0), minute(), secondIndicator());
		PRINTLN(buffer);
		display.setChars(buffer);
		oldTime = currentTime;
	}
	display.refreshDisplay(); 
}

// Retrieve seconds, period or emoty
char* secondIndicator(){
	static char s[2];
	s[0] = 0;
	
	if (showSeconds && (second() % 2 == 0))
		strcpy(s, ".");

	return s;
}

// Run display and wait for a button to be pressed
void displayAndWaitForButton(long wait, Button b){
	unsigned long finish = millis() + wait;
	while (millis() < finish){
		display.refreshDisplay();
		if (!b.isPressed())
			break;
	}
}

// Run display and wait
void clearAndWait(){
	display.blank();
	unsigned long finish = millis() + CLEAR_DELAY_MS;
	while (millis() < finish){
		display.refreshDisplay();
	}
}

// Run display and wait
void displayAndWait(long wait){
	unsigned long finish = millis() + wait;
	while (millis() < finish)
		display.refreshDisplay();
}

// Scroll a pretty message accross the screen
void scrollMessage(char* msg){
	if (strlen(msg) == 0) return;
	static char s[NUMBER_OF_DIGITS + 1];

	// Add some spaces to allow scrolling from left pretty
	memset(s, ' ', NUMBER_OF_DIGITS); 
	s[NUMBER_OF_DIGITS] = 0;
	display.blank();
	displayAndWait(SCROLL_DELAY_MS);

	// Display string appearing from left
	for (int8_t i = NUMBER_OF_DIGITS - 1; i > 0; i--){
		strncpy(s + i, msg, NUMBER_OF_DIGITS - i);
		display.setChars(s);
		displayAndWait(SCROLL_DELAY_MS);
	}

	// Loop through entire string + null (not displayed)
	for (char* p = msg; p <= msg + strlen(msg); p++){
		display.setChars(p);
		displayAndWait(SCROLL_DELAY_MS);
	}
}

// Short scroll with no spaces padding
void shortScroll(char* msg){
	for (char* p = msg; p <= msg + strlen(msg) - NUMBER_OF_DIGITS; p++){
		display.setChars(p);
		displayAndWait(SHORT_SCROLL_DELAY_MS);
	}
}

// Get the number of days in a month
uint8_t getDaysInMonth(uint8_t m){
	uint8_t leaper = (m == 2 && isLeapYear() ? 1 : 0 );
	return ((m > 0) && (m < 13)) ? daysInMonth[m - 1] + leaper : 1;
}

// Is it a leap year
bool isLeapYear(){
	if (year() % 4 == 0) {
		if (year() % 100 != 0)
			return true;
		else 
			if (year() % 400 == 0)
				return true;
	}
	return false;
}

// Initially EEPROM and mark time as adjusted (OK)
void initEEPROM(){
	if (EEPROM.read(ADDR_FLAG) != INITIALISED){
		EEPROM.write(ADDR_FLAG, INITIALISED);

		// If we're in DST set flag so the time doesn't got adjusted
		timeConfirmed();
	}
	else{
		PRINTLNF("Reading EEPROM");
		readShowSeconds();
		readDSTAdjusted();
		readFormat();
	}
}

// If the user has updated the time manually then we do not wish to automatically change the time
void timeConfirmed(){
	saveDSTadjustment(isDST());
}

// Run our DST check alright
bool checkDST(){
	if (isDST() && !dstAdjusted){
		// Enable dst
		PRINTLNF("DST adjustment +hour!");
		adjustTime(true);
		return true;
	}
	else if (!isDST() && dstAdjusted){
		// Disable DST
		PRINTLNF("DST adjustment -hour!");
		adjustTime(false);
		return true;
	}
	return false;
}

// Adjust time according to DST pattern
void adjustTime(bool dst){
	uint8_t newHour = hour();
	
	// Adjust hour
	if (dst)
		newHour++;
	else
		newHour--;

	// Store the new time
	storeTime(newHour, minute(), second(), day(), month(), year());

	// Store the fact we've adjusted the time
	saveDSTadjustment(dst);
}

// Store the fact we've adjusted time for DST change
void saveDSTadjustment(bool adjusted){
	dstAdjusted = adjusted;
	EEPROM.write(ADDR_DST, adjusted);
	PRINTVAR("EEPROM: DST adjustment flag set: ", adjusted);
}

// Save showing seconds
void saveShowSeconds(bool show){
	showSeconds = show;
	EEPROM.write(ADDR_SECONDS, show);
	PRINTVAR("EEPROM: show seconds set: ", show);
}

// Read showing sceonds
void readShowSeconds(){
	showSeconds = EEPROM.read(ADDR_SECONDS);
	PRINTVAR("Show seconds? " , showSeconds);
}

// Save showing seconds
void saveFormat(bool twentyFourHours){
	twentyFourHour = twentyFourHours;
	EEPROM.write(ADDR_FORMAT, twentyFourHour);
	PRINTVAR("EEPROM: 24 hour format set: ", twentyFourHour);
}

// Read showing sceonds
void readFormat(){
	twentyFourHour = EEPROM.read(ADDR_FORMAT);
	PRINTVAR("24 hour format? " , twentyFourHour);
}

// Read from EEPROM if we have adjusted for DST already
bool readDSTAdjusted(){
	dstAdjusted = EEPROM.read(ADDR_DST);
	PRINTVAR("DST adjustment? " , dstAdjusted);
}

// This is for the UK, people!
//
// http://my-small-projects.blogspot.co.uk/2015/05/arduino-checking-for-british-summer-time.html
bool isDST()
{
	int imonth = month();
	int iday = day();
	int hr = hour();
	
	//January, february, and november are out.
	if (imonth < 3 || imonth > 10) { return false; }
	//April to September are in
	if (imonth > 3 && imonth < 10) { return true; }

	// find last sun in mar and oct - quickest way I've found to do it
	// last sunday of march
	int lastMarSunday =  (31 - (5* year() /4 + 4) % 7);
	//last sunday of october
	int lastOctSunday = (31 - (5 * year() /4 + 1) % 7);
			
	//In march, we are BST if is the last sunday in the month
	if (imonth == 3) { 
		
		if( iday > lastMarSunday)
			return true;
		if( iday < lastMarSunday)
			return false;
		
		if (hr < 1)
			return false;
						
		return true; 

	}
	//In October we must be before the last sunday to be bst.
	//That means the previous sunday must be before the 1st.
	if (imonth == 10) { 

		if( iday < lastOctSunday)
			return true;
		if( iday > lastOctSunday)
			return false;  
		
		if (hr >= 1)
			return false;
			
		return true;  
	}
}